# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest
from helpers import IDCollector

import npyosmium


def test_tag_filter_no_keys():
    with pytest.raises(TypeError, match="tags to filter"):
        npyosmium.filter.TagFilter()


@pytest.mark.parametrize('kv', ['something', 'so', ('a', 'b', 'c'),
                                (None, 'a'), ('a', None),
                                (34, 'a'), ('a', 0)])
def test_tag_filter_bad_arguments(kv):
    with pytest.raises(TypeError, match="Each tag must be a tuple"):
        npyosmium.filter.TagFilter(kv)


@pytest.mark.parametrize('tags,filt', [('foo=bar', [('foo', 'bar')]),
                                       ('a=1,b=2', [('x', 'x'), ('a', '1')]),
                                       ('a=1,b=2', [('x', 'x'), ('b', '2')])
                                       ])
def test_tag_filter_pass(opl_reader, tags, filt):
    data = f"""\
            n1 T{tags}
            w2 T{tags}
            r3 T{tags}
            c4 T{tags}
            """

    ids = IDCollector()

    npyosmium.apply(opl_reader(data), npyosmium.filter.TagFilter(*filt), ids)

    assert ids.nodes == [1]
    assert ids.ways == [2]
    assert ids.relations == [3]
    assert ids.changesets == [4]


@pytest.mark.parametrize('tags,filt', [('foo=bar', [('foo', 'bars')]),
                                       ('a=1,b=2', [('x', 'x'), ('a', '2')])
                                       ])
def test_tag_filter_fail(opl_reader, tags, filt):
    data = f"""\
            n1 T{tags}
            w2 T{tags}
            r3 T{tags}
            c4 T{tags}
            """

    ids = IDCollector()

    npyosmium.apply(opl_reader(data), npyosmium.filter.TagFilter(*filt), ids)

    assert ids.nodes == []
    assert ids.ways == []
    assert ids.relations == []
    assert ids.changesets == []



"""
Simple example that counts the number of changes in an osm diff file.

Shows how to detect the different kind of modifications and how to
use the handler generator function instead of a handler class.
"""
import sys

import npyosmium


class Stats:

    def __init__(self):
        self.added = 0
        self.modified = 0
        self.deleted = 0

    def add(self, o):
        if o.deleted:
            self.deleted += 1
        elif o.version == 1:
            self.added += 1
        else:
            self.modified += 1

    def outstats(self, prefix):
        print(f"{prefix} added: {self.added}")
        print(f"{prefix} modified: {self.modified}")
        print(f"{prefix} deleted: {self.deleted}")


def main(osmfile):
    stats = {t: Stats() for t in 'nwr'}

    for obj in npyosmium.FileProcessor(osmfile):
        stats[obj.type_str()].add(obj)

    stats['n'].outstats("Nodes")
    stats['w'].outstats("Ways")
    stats['r'].outstats("Relations")

    return 0


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage: python %s <osmfile>" % sys.argv[0])
        sys.exit(-1)

    exit(main(sys.argv[1]))



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest
from helpers import IDCollector

import npyosmium


def test_filter_no_keys():
    with pytest.raises(TypeError, match="keys to filter"):
        npyosmium.filter.KeyFilter()


@pytest.mark.parametrize('key', [None, 1, IDCollector(), 'a'.encode('utf-8')])
def test_filter_bad_argument_types(key):
    with pytest.raises(TypeError, match="must be strings"):
        npyosmium.filter.KeyFilter("foo", key)


@pytest.mark.parametrize('key,nodes,changesets', [('foo', [1], [10]),
                                                  ('name', [1, 2], [20]),
                                                  ('kö*', [4], [])])
def test_filter_simple(opl_reader, key, nodes, changesets):
    data = """\
           n1 Tfoo=bar,name=loo
           n2 Tname=else
           n3 x9 y0
           n4 Tkö*=fr
           c10 Tfoo=baz
           c20 Tname=none
           """

    post = IDCollector()

    npyosmium.apply(opl_reader(data), npyosmium.filter.KeyFilter(key), post)

    assert post.nodes == nodes
    assert post.changesets == changesets



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import ClassVar, Union, overload

import npyosmium.osm

ALL: use_nodes
BACKWARD: direction
FORWARD: direction
UNIQUE: use_nodes

PointLike = Union[npyosmium.osm.Node, npyosmium.osm.Location, npyosmium.osm.NodeRef]
LineStringLike = Union[npyosmium.osm.Way, npyosmium.osm.WayNodeList]

class use_nodes:
    ALL: ClassVar[use_nodes] = ...
    UNIQUE: ClassVar[use_nodes] = ...
    def __init__(self, value: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class direction:
    BACKWARD: ClassVar[direction] = ...
    FORWARD: ClassVar[direction] = ...
    def __init__(self, value: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class Coordinates:
    """ Represent a x/y coordinate. The projection of the coordinate
        is left to the interpretation of the caller.
    """
    @overload
    def __init__(self) -> None:
        """ Create an invalid coordinate.
        """
    @overload
    def __init__(self, cx: float, cy: float) -> None:
        """ Create a coordinate with the given x and y values.
        """
    @overload
    def __init__(self, location: npyosmium.osm.Location) -> None:
        """ Create a coordinate from a node location.
        """
    def valid(self) -> bool:
        """ Return true if the coordinate is valid.
            A coordinate can only be invalid when both x and y are NaN.
        """
    @property
    def x(self) -> float:
        """ x portion of the coordinate.
        """
    @property
    def y(self) -> float:
        """ y portion of the coordinate.
        """

class FactoryProtocol:
    """ Protocol for classes that implement the necessary functions
        for converting OSM objects into simple-feature-like geometries.
    """
    def create_linestring(self, line: LineStringLike, use_nodes: use_nodes = ..., direction: direction = ...) -> str:
        """ Create a line string geometry from a way like object. This may be a
            [Way][npyosmium.osm.Way] or a [WayNodeList][npyosmium.osm.WayNodeList].
            Subsequent nodes with the exact same coordinates will be filtered
            out because many tools consider repeated coordinates in a
            line string invalid. Set _use_nodes_ to `npyosmium.geom.ALL` to
            suppress this behaviour.

            The line string usually follows the order of the node list.
            Set _direction_ to `npyosmium.geom.BACKWARDS` to inverse the direction.
        """
    def create_multipolygon(self, area: npyosmium.osm.Area) -> str:
        """ Create a multi-polygon geometry from an [Area][npyosmium.osm.Area]
            object.
        """
    def create_point(self, location: PointLike) -> str:
        """ Create a point geometry from a [Node][npyosmium.osm.Node], a
            location or a node reference.
        """
    @property
    def epsg(self) -> int:
        """ Projection of the output geometries as a EPSG number.
        """
    @property
    def proj_string(self) -> str:
        """ Projection of the output geometries as a projection string.
        """

class GeoJSONFactory(FactoryProtocol):
    """ Factory that creates GeoJSON geometries from npyosmium geometries.
    """

class WKBFactory(FactoryProtocol):
    """ Factory that creates WKB from npyosmium geometries.
    """

class WKTFactory(FactoryProtocol):
    pass

def haversine_distance(list: npyosmium.osm.WayNodeList) -> float: ...
def lonlat_to_mercator(coordinate: Coordinates) -> Coordinates: ...
def mercator_to_lonlat(coordinate: Coordinates) -> Coordinates: ...



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
""" Helper functions for change file handling. """
import datetime as dt
import logging
from typing import NamedTuple, Optional

from npyosmium.io import Reader as oreader
from npyosmium.osm import NOTHING

LOG = logging.getLogger('npyosmium')


class ReplicationHeader(NamedTuple):
    """ Description of a replication state.
    """
    url: Optional[str]
    "Base URL of the replication service."
    sequence: Optional[int]
    "ID of the change file on the server."
    timestamp: Optional[dt.datetime]
    "Date of latest changes contained in the diff file."


def get_replication_header(fname: str) -> ReplicationHeader:
    """ Scans the given file for an Osmosis replication header. It returns
        a namedtuple with `url`, `sequence` and `timestamp`. Each or all fields
        may be None, if the piece of information is not avilable. If any of
        the fields has an invalid format, it is simply ignored.

        The given file must exist and be readable for osmium, otherwise
        a `RuntimeError` is raised.
    """

    r = oreader(fname, NOTHING)
    h = r.header()

    tsstr = h.get("osmosis_replication_timestamp")
    url: Optional[str] = h.get("osmosis_replication_base_url")

    if url or tsstr:
        LOG.debug("Replication information found in OSM file header.")

    if url:
        LOG.debug("Replication URL: %s", url)
        # the sequence ID is only considered valid, if an URL is given
        seqstr = h.get("osmosis_replication_sequence_number")
        seq: Optional[int]
        if seqstr:
            LOG.debug("Replication sequence: %s", seqstr)
            try:
                seq = int(seqstr)
                if seq < 0:
                    LOG.warning("Sequence id '%d' in OSM file header is negative. Ignored.", seq)
                    seq = None
            except ValueError:
                LOG.warning("Sequence id '%s' in OSM file header is not a number. Ignored.", seq)
                seq = None
        else:
            seq = None
    else:
        url = None
        seq = None

    if tsstr:
        LOG.debug("Replication timestamp: %s", tsstr)
        try:
            ts = dt.datetime.strptime(tsstr, "%Y-%m-%dT%H:%M:%SZ")
            ts = ts.replace(tzinfo=dt.timezone.utc)

        except ValueError:
            LOG.warning("Date in OSM file header is not in ISO8601 format"
                        "(e.g. 2015-12-24T08:08Z). Ignored.")
            ts = None
    else:
        ts = None

    return ReplicationHeader(url, seq, ts)



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import pytest
from helpers import IDCollector

import npyosmium


@pytest.fixture
def reader(opl_reader):
    def _mk():
        return opl_reader("""\
               n1 x1 y1
               n2 x1 y1 Tfoo=bar
               w1 Nn1,n2 Thighway=road
               w34 Nn2,n1
               r90
               r91 Tsome=thing
               c222 Ttodo=done
               c223
               """)

    return _mk


def test_filter_default_config(reader):
    pre = IDCollector()
    post = IDCollector()
    npyosmium.apply(reader(), pre, npyosmium.filter.EmptyTagFilter(), post)

    assert pre.nodes == [1, 2]
    assert post.nodes == [2]
    assert pre.ways == [1, 34]
    assert post.ways == [1]
    assert pre.relations == [90, 91]
    assert post.relations == [91]
    assert pre.changesets == [222, 223]
    assert post.changesets == [222]


def test_filter_restrict_entity(reader):
    pre = IDCollector()
    post = IDCollector()
    npyosmium.apply(reader(), pre,
                 npyosmium.filter.EmptyTagFilter().enable_for(npyosmium.osm.WAY | npyosmium.osm.RELATION),
                 post)

    assert pre.nodes == [1, 2]
    assert post.nodes == [1, 2]
    assert pre.ways == [1, 34]
    assert post.ways == [1]
    assert pre.relations == [90, 91]
    assert post.relations == [91]


def test_filter_chained(reader):
    pre = IDCollector()
    post = IDCollector()
    npyosmium.apply(reader(), pre,
                 npyosmium.filter.EmptyTagFilter().enable_for(npyosmium.osm.NODE),
                 npyosmium.filter.EmptyTagFilter().enable_for(npyosmium.osm.WAY),
                 post)

    assert pre.nodes == [1, 2]
    assert post.nodes == [2]
    assert pre.ways == [1, 34]
    assert post.ways == [1]
    assert pre.relations == [90, 91]
    assert post.relations == [90, 91]



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
""" Provides some helper functions for test.
"""
from datetime import datetime, timezone

import npyosmium


def mkdate(*args):
    return datetime(*args, tzinfo=timezone.utc)


class CountingHandler(npyosmium.SimpleHandler):

    def __init__(self):
        super(CountingHandler, self).__init__()
        self.counts = [0, 0, 0, 0]

    def node(self, _):
        self.counts[0] += 1

    def way(self, _):
        self.counts[1] += 1

    def relation(self, _):
        self.counts[2] += 1

    def area(self, _):
        self.counts[3] += 1


class IDCollector:

    def __init__(self):
        self.nodes = []
        self.ways = []
        self.relations = []
        self.changesets = []

    def node(self, n):
        self.nodes.append(n.id)

    def way(self, w):
        self.ways.append(w.id)

    def relation(self, r):
        self.relations.append(r.id)

    def changeset(self, c):
        self.changesets.append(c.id)



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
"""
Typing information for generated version module.
"""

pyosmium_release: str
libosmium_version: str
protozero_version: str
pybind11_version: str



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import uuid
from collections import OrderedDict
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone

import pytest
from helpers import mkdate

import npyosmium


@pytest.fixture
def test_writer(tmp_path):
    @contextmanager
    def _WriteExpect(filename, expected):
        with npyosmium.SimpleWriter(str(filename), 1024*1024) as writer:
            yield writer

        assert filename.read_text().strip() == expected

    def _create(expected):
        filename = tmp_path / f"{uuid.uuid4()}.opl"
        return _WriteExpect(filename, expected)

    return _create


class O:  # noqa: E742
    def __init__(self, **params):
        for k, v in params.items():
            setattr(self, k, v)


@pytest.mark.parametrize('osmobj, out_attr', [
      (O(id=None), '0 v0 dV c0 t i0 u T'),
      (O(visible=None), '0 v0 dV c0 t i0 u T'),
      (O(version=None), '0 v0 dV c0 t i0 u T'),
      (O(uid=None), '0 v0 dV c0 t i0 u T'),
      (O(user=None), '0 v0 dV c0 t i0 u T'),
      (O(timestamp=None), '0 v0 dV c0 t i0 u T'),
      (O(id=1), '1 v0 dV c0 t i0 u T'),
      (O(id=-99), '-99 v0 dV c0 t i0 u T'),
      (O(visible=True), '0 v0 dV c0 t i0 u T'),
      (O(visible=False), '0 v0 dD c0 t i0 u T'),
      (O(version=23), '0 v23 dV c0 t i0 u T'),
      (O(user="Schmidt"), '0 v0 dV c0 t i0 uSchmidt T'),
      (O(user=""), '0 v0 dV c0 t i0 u T'),
      (O(uid=987), '0 v0 dV c0 t i987 u T'),
      (O(timestamp='2012-04-14T20:58:35Z'), '0 v0 dV c0 t2012-04-14T20:58:35Z i0 u T'),
      (O(timestamp=mkdate(2009, 4, 14, 20, 58, 35)), '0 v0 dV c0 t2009-04-14T20:58:35Z i0 u T'),
      (O(timestamp=datetime(2009, 4, 14, 20, 58, 35, tzinfo=timezone(timedelta(hours=1)))),
       '0 v0 dV c0 t2009-04-14T19:58:35Z i0 u T'),
      (O(timestamp='1970-01-01T00:00:01Z'), '0 v0 dV c0 t1970-01-01T00:00:01Z i0 u T')
    ])
class TestWriteAttributes:
    def test_node_simple_attr(self, test_writer, osmobj, out_attr):
        with test_writer('n' + out_attr + ' x y') as w:
            w.add_node(osmobj)

    def test_way_simple_attr(self, test_writer, osmobj, out_attr):
        with test_writer('w' + out_attr + ' N') as w:
            w.add_way(osmobj)

    def test_relation_simple_attr(self, test_writer, osmobj, out_attr):
        with test_writer('r' + out_attr + ' M') as w:
            w.add_relation(osmobj)


@pytest.mark.parametrize('tags,out', [
     (None, 'T'),
     ([], 'T'),
     ({}, 'T'),
     ((("foo", "bar"), ), 'Tfoo=bar'),
     ((("foo", "bar"), ("2", "1")), 'Tfoo=bar,2=1'),
     ({'test': 'drive'}, 'Ttest=drive'),
     (OrderedDict((('a', 'b'), ('c', '3'))), 'Ta=b,c=3'),
    ])
class TestWriteTags:
    def test_node_tags(self, test_writer, tags, out):
        with test_writer('n0 v0 dV c0 t i0 u ' + out + ' x y') as w:
            w.add_node(O(tags=tags))

    def test_way_tags(self, test_writer, tags, out):
        with test_writer('w0 v0 dV c0 t i0 u ' + out + ' N') as w:
            w.add_way(O(tags=tags))

    def test_relation_tags(self, test_writer, tags, out):
        with test_writer('r0 v0 dV c0 t i0 u ' + out + ' M') as w:
            w.add_relation(O(tags=tags))


@pytest.mark.parametrize("location,out", [((1.1234561, 0.1234561), 'x1.1234561 y0.1234561'),
                                          ((30.46, 50.37), 'x30.46 y50.37'),
                                          (None, 'x y')])
def test_location(test_writer, location, out):
    with test_writer('n0 v0 dV c0 t i0 u T ' + out) as w:
        w.add_node(O(location=location))

def test_add_locations(test_writer):
    # numpy was not a dependency of older shapely version used with python 3.6
    numpy = pytest.importorskip("numpy", reason="numpy is not available")
    locations = numpy.array([[1.,1.1], [2.,2.2], [3.,3.3], [4.,4.4], [4.,4.4]])
    expected = """n1000 v0 dV c0 t i0 u T x1 y1.1
n1001 v0 dV c0 t i0 u T x2 y2.2
n1002 v0 dV c0 t i0 u T x3 y3.3
n1003 v0 dV c0 t i0 u T x4 y4.4
n1004 v0 dV c0 t i0 u T x4 y4.4"""
    with test_writer(expected) as w:
        w.add_locations(locations=locations, first_id=1000)

def test_location_generic(test_writer):
    with test_writer('n0 v0 dV c0 t i0 u T x30.46 y50.37') as w:
        w.add_node(O(location=(30.46, 50.37)))


def test_node_list(test_writer):
    with test_writer('w0 v0 dV c0 t i0 u T Nn1,n2,n3,n-4') as w:
        w.add_way(O(nodes=(1, 2, 3, -4)))


def test_node_list_generic(test_writer):
    with test_writer('w0 v0 dV c0 t i0 u T Nn1,n2,n3,n-4') as w:
        w.add(O(nodes=(1, 2, 3, -4)))


def test_node_list_none(test_writer):
    with test_writer('w0 v0 dV c0 t i0 u T N') as w:
        w.add_way(O(nodes=None))


def test_relation_members(test_writer):
    with test_writer('r0 v0 dV c0 t i0 u T Mn34@foo,r200@,w1111@x') as w:
        w.add_relation(O(members=(('n', 34, 'foo'),
                                  ('r', 200, ''),
                                  ('w', 1111, 'x')
                                  )))


def test_relation_members_generic(test_writer):
    with test_writer('r0 v0 dV c0 t i0 u T Mn34@foo,r200@,w1111@x') as w:
        w.add(O(members=(('n', 34, 'foo'),
                         ('r', 200, ''),
                         ('w', 1111, 'x')
                         )))


def test_relation_members_None(test_writer):
    with test_writer('r0 v0 dV c0 t i0 u T M') as w:
        w.add_relation(O(members=None))


def test_node_object(test_writer, simple_handler):
    node_opl = 'n235 v1 dV c0 t i0 u Telephant=yes x98.7 y-3.45'

    with test_writer(node_opl) as w:
        simple_handler(node_opl, node=lambda o: w.add_node(o))


def test_node_object_generic(test_writer, simple_handler):
    node_opl = 'n235 v1 dV c0 t i0 u Telephant=yes x98.7 y-3.45'

    with test_writer(node_opl) as w:
        simple_handler(node_opl, node=lambda o: w.add(o))


def test_location_object(test_writer, simple_handler):
    node_opl = 'n235 v1 dV c0 t i0 u Telephant=yes x98.7 y-3.45'

    with test_writer('n0 v0 dV c0 t i0 u T x98.7 y-3.45') as w:
        simple_handler(node_opl, node=lambda o: w.add_node(O(location=o.location)))


def test_tag_object(test_writer, simple_handler):
    node_opl = 'n235 v1 dV c0 t i0 u Telephant=yes x98.7 y-3.45'

    with test_writer('n0 v0 dV c0 t i0 u Telephant=yes x y') as w:
        simple_handler(node_opl, node=lambda o: w.add_node(O(tags=o.tags)))


def test_way_object(test_writer, simple_handler):
    way_opl = 'w45 v14 dV c0 t i0 u Thighway=top Nn23,n56,n34,n23'

    with test_writer(way_opl) as w:
        simple_handler(way_opl, way=lambda o: w.add_way(o))


def test_way_object_generic(test_writer, simple_handler):
    way_opl = 'w45 v14 dV c0 t i0 u Thighway=top Nn23,n56,n34,n23'

    with test_writer(way_opl) as w:
        simple_handler(way_opl, way=lambda o: w.add(o))


def test_nodelist_object(test_writer, simple_handler):
    way_opl = 'w45 v14 dV c0 t i0 u Thighway=top Nn23,n56,n34,n23'

    with test_writer('w0 v0 dV c0 t i0 u T Nn23,n56,n34,n23') as w:
        simple_handler(way_opl, way=lambda o: w.add_way(O(nodes=o.nodes)))


def test_noderef_object(test_writer, simple_handler):
    way_opl = 'w45 v14 dV c0 t i0 u Thighway=top Nn23,n56,n34,n23'

    with test_writer('w0 v0 dV c0 t i0 u T Nn56,n34') as w:
        simple_handler(way_opl,
                       way=lambda o: w.add_way(O(nodes=[n for n in o.nodes if n.ref != 23])))


def test_relation_object(test_writer, simple_handler):
    rel_opl = 'r2 v0 dV c0 t i0 u Ttype=multipolygon Mw1@,w2@,w3@inner'

    with test_writer(rel_opl) as w:
        simple_handler(rel_opl, relation=lambda o: w.add_relation(o))


def test_relation_object_generic(test_writer, simple_handler):
    rel_opl = 'r2 v0 dV c0 t i0 u Ttype=multipolygon Mw1@,w2@,w3@inner'

    with test_writer(rel_opl) as w:
        simple_handler(rel_opl, relation=lambda o: w.add(o))


def test_memberlist_object(test_writer, simple_handler):
    rel_opl = 'r2 v0 dV c0 t i0 u Ttype=multipolygon Mw1@,w2@,w3@inner'

    with test_writer('r0 v0 dV c0 t i0 u T Mw1@,w2@,w3@inner') as w:
        simple_handler(rel_opl,
                       relation=lambda o: w.add_relation(O(members=o.members)))


def test_member_object(test_writer, simple_handler):
    rel_opl = 'r2 v0 dV c0 t i0 u Ttype=multipolygon Mw1@,w2@,w3@inner'

    with test_writer('r0 v0 dV c0 t i0 u T Mw1@,w2@') as w:
        simple_handler(rel_opl,
                       relation=lambda o: w.add_relation(O(members=[m for m in o.members
                                                                    if m.role != 'inner'])))


def test_set_custom_header(tmp_path):
    fn = tmp_path / f"{uuid.uuid4()}.xml"
    h = npyosmium.io.Header()
    h.set('generator', 'foo')
    h.add_box(npyosmium.osm.Box(0.1, -4, 10, 45))

    writer = npyosmium.SimpleWriter(fn, 4000, h)

    try:
        writer.add_node({})
    finally:
        writer.close()

    with npyosmium.io.Reader(fn) as rd:
        h = rd.header()
        assert h.get('generator') == 'foo'
        assert h.box().valid()
        assert h.box().bottom_left == npyosmium.osm.Location(0.1, -4)
        assert h.box().top_right == npyosmium.osm.Location(10, 45)


def test_add_node_after_close(tmp_path, simple_handler):
    node_opl = "n235 v1 dV c0 t i0 u Telephant=yes x98.7 y-3.45"

    filename = tmp_path / f"{uuid.uuid4()}.opl"
    writer = npyosmium.SimpleWriter(str(filename), 1024*1024)
    writer.close()

    with pytest.raises(RuntimeError, match='closed'):
        simple_handler(node_opl, node=lambda o: writer.add_node(o))


def test_add_way_after_close(tmp_path, simple_handler):
    node_opl = "w1 Nn1"

    filename = tmp_path / f"{uuid.uuid4()}.opl"
    writer = npyosmium.SimpleWriter(str(filename), 1024*1024)
    writer.close()

    with pytest.raises(RuntimeError, match='closed'):
        simple_handler(node_opl, way=lambda o: writer.add_way(o))


def test_add_relation_after_close(tmp_path, simple_handler):
    node_opl = "r54 Mn1@,w3@foo"

    filename = tmp_path / f"{uuid.uuid4()}.opl"
    writer = npyosmium.SimpleWriter(str(filename), 1024*1024)
    writer.close()

    with pytest.raises(RuntimeError, match='closed'):
        simple_handler(node_opl, relation=lambda o: writer.add_relation(o))


@pytest.mark.parametrize("final_item", (True, False))
def test_catch_errors_in_add_node(tmp_path, final_item):
    test_file = tmp_path / f"{uuid.uuid4()}.opl"

    with npyosmium.SimpleWriter(str(test_file), 4000) as writer:
        writer.add_node(npyosmium.osm.mutable.Node(id=123))
        with pytest.raises(TypeError):
            writer.add_node(npyosmium.osm.mutable.Node(id=124, tags=34))
        if not final_item:
            writer.add_node(npyosmium.osm.mutable.Node(id=125))

    output = test_file.read_text()

    expected = 'n123 v0 dV c0 t i0 u T x y\n'
    if not final_item:
        expected += 'n125 v0 dV c0 t i0 u T x y\n'

    assert output == expected


@pytest.mark.parametrize("final_item", (True, False))
def test_catch_errors_in_add_way(tmp_path, final_item):
    test_file = tmp_path / f"{uuid.uuid4()}.opl"

    with npyosmium.SimpleWriter(test_file, 4000) as writer:
        writer.add_way(npyosmium.osm.mutable.Way(id=123, nodes=[1, 2, 3]))
        with pytest.raises(TypeError):
            writer.add_way(npyosmium.osm.mutable.Way(id=124, nodes=34))
        if not final_item:
            writer.add_way(npyosmium.osm.mutable.Way(id=125, nodes=[11, 12]))

    output = test_file.read_text()

    expected = 'w123 v0 dV c0 t i0 u T Nn1,n2,n3\n'
    if not final_item:
        expected += 'w125 v0 dV c0 t i0 u T Nn11,n12\n'

    assert output == expected


@pytest.mark.parametrize("final_item", (True, False))
def test_catch_errors_in_add_relation(tmp_path, final_item):
    test_file = tmp_path / f"{uuid.uuid4()}.opl"

    with npyosmium.SimpleWriter(filename=str(test_file), bufsz=4000) as writer:
        writer.add_relation(npyosmium.osm.mutable.Relation(id=123))
        with pytest.raises(TypeError):
            writer.add_relation(npyosmium.osm.mutable.Relation(id=124, members=34))
        if not final_item:
            writer.add_relation(npyosmium.osm.mutable.Relation(id=125))

    output = test_file.read_text()

    expected = 'r123 v0 dV c0 t i0 u T M\n'
    if not final_item:
        expected += 'r125 v0 dV c0 t i0 u T M\n'

    assert output == expected


def test_do_not_overwrite_by_default(tmp_path):
    test_file = tmp_path / f"{uuid.uuid4()}.opl"

    with npyosmium.SimpleWriter(filename=str(test_file), bufsz=4000) as writer:
        writer.add_node(npyosmium.osm.mutable.Node(id=123))

    # try to open again
    with pytest.raises(RuntimeError, match='Open failed'):
        npyosmium.SimpleWriter(filename=str(test_file))


def test_do_overwrite(tmp_path):
    test_file = tmp_path / f"{uuid.uuid4()}.opl"

    with npyosmium.SimpleWriter(filename=str(test_file), bufsz=4000) as writer:
        writer.add_node(npyosmium.osm.mutable.Node(id=123))

    with npyosmium.SimpleWriter(filename=str(test_file), overwrite=True) as writer:
        pass


def test_write_to_file(tmp_path):
    test_file = tmp_path / f"{uuid.uuid4()}.txt"

    with npyosmium.SimpleWriter(npyosmium.io.File(test_file, 'opl'), bufsz=4000) as writer:
        writer.add_node(npyosmium.osm.mutable.Node(id=123))



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import threading

from helpers import CountingHandler


def test_threaded_processing():
    """ Process a file in a different thread and make sure that processing
        completes.
    """
    function_complete = threading.Event()
    c = CountingHandler()

    def import_data():
        c.apply_buffer('n1 x67.8 y-45.6932'.encode('utf-8'), 'opl')
        function_complete.set()

    t = threading.Thread(target=import_data)
    t.start()
    function_complete.wait(timeout=2)

    assert function_complete.is_set()
    assert c.counts == [1, 0, 0, 0]



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
""" Tests for the npyosmium-get-changes script.
"""
import datetime as dt
import uuid
from textwrap import dedent

import pytest
from helpers import IDCollector

import npyosmium
from npyosmium.tools.pyosmium_get_changes import pyosmium_get_changes

try:
    import http.cookiejar as cookiejarlib
except ImportError:
    import cookielib as cookiejarlib


REPLICATION_BASE_TIME = dt.datetime(year=2017, month=8, day=26, hour=11, tzinfo=dt.timezone.utc)
REPLICATION_BASE_SEQ = 100
REPLICATION_CURRENT = 140


@pytest.fixture
def replication_server(httpserver):
    def _state(seq):
        seqtime = REPLICATION_BASE_TIME + dt.timedelta(hours=seq - REPLICATION_CURRENT)
        timestamp = seqtime.strftime('%Y-%m-%dT%H\\:%M\\:%SZ')
        return f"sequenceNumber={seq}\ntimestamp={timestamp}\n"

    httpserver.no_handler_status_code = 404
    httpserver.expect_request('/state.txt').respond_with_data(_state(REPLICATION_CURRENT))
    for i in range(REPLICATION_BASE_SEQ, REPLICATION_CURRENT + 1):
        httpserver.expect_request(f'/000/000/{i}.opl')\
                  .respond_with_data(f"r{i} M" + ",".join(f"n{i}@" for i in range(1, 6000)))
        httpserver.expect_request(f'/000/000/{i}.state.txt').respond_with_data(_state(i))

    return httpserver.url_for('')


@pytest.fixture
def runner(httpserver):
    def _run(*args):
        return pyosmium_get_changes(
            ['--server', httpserver.url_for(''), '--diff-type', 'opl'] + list(map(str, args)))

    return _run


def test_init_id(runner, capsys, replication_server):
    assert 0 == runner('-I', '100')

    output = capsys.readouterr().out.strip()

    assert output == '100'


def test_init_date(runner, capsys, httpserver):
    httpserver.expect_request('/state.txt').respond_with_data(dedent("""\
                sequenceNumber=100
                timestamp=2017-08-26T11\\:04\\:02Z
             """))
    httpserver.expect_request('/000/000/000.state.txt').respond_with_data(dedent("""\
                sequenceNumber=0
                timestamp=2016-08-26T11\\:04\\:02Z
             """))
    assert 0 == runner('-D', '2015-12-24T08:08:08Z')

    output = capsys.readouterr().out.strip()

    assert output == '-1'


def test_init_to_file(runner, tmp_path, replication_server):
    fname = tmp_path / f"{uuid.uuid4()}.seq"

    assert 0 == runner('-I', '130', '-f', fname)
    assert fname.read_text() == '130'


def test_init_from_seq_file(runner, tmp_path, replication_server):
    fname = tmp_path / f"{uuid.uuid4()}.seq"
    fname.write_text('140')

    assert 0 == runner('-f', fname)
    assert fname.read_text() == '140'


def test_init_date_with_cookie(runner, capsys, tmp_path, httpserver):
    httpserver.expect_request('/state.txt').respond_with_data(dedent("""\
                sequenceNumber=100
                timestamp=2017-08-26T11\\:04\\:02Z
             """))
    httpserver.expect_request('/000/000/000.state.txt').respond_with_data(dedent("""\
                sequenceNumber=0
                timestamp=2016-08-26T11\\:04\\:02Z
             """))

    fname = tmp_path / 'my.cookie'
    cookie_jar = cookiejarlib.MozillaCookieJar(str(fname))
    cookie_jar.save()

    assert 0 == runner('--cookie', fname, '-D', '2015-12-24T08:08:08Z')

    output = capsys.readouterr().out.strip()

    assert output == '-1'


def test_get_simple_update(runner, tmp_path, replication_server):
    outfile = tmp_path / f"{uuid.uuid4()}.opl"

    assert 0 == runner('-I', '139', '-o', outfile)

    ids = IDCollector()
    npyosmium.apply(outfile, ids)

    assert ids.nodes == []
    assert ids.ways == []
    assert ids.relations == [140]


@pytest.mark.parametrize('end_id,max_size,actual_end', [(107, None, 107),
                                                        (None, 1, 108),
                                                        (105, 1, 105),
                                                        (110, 1, 108)])
def test_apply_diffs_endid(runner, tmp_path, replication_server, end_id, max_size, actual_end):
    outfile = tmp_path / f"{uuid.uuid4()}.opl"

    params = ['-I', '100', '-o', outfile]
    if end_id is not None:
        params.extend(('--end-id', end_id))
    if max_size is not None:
        params.extend(('-s', max_size))

    assert 0 == runner(*params)

    ids = IDCollector()
    npyosmium.apply(str(outfile), ids)

    assert ids.relations == list(range(101, actual_end + 1))


def test_change_id_too_old_for_replication_source(runner, tmp_path, replication_server, caplog):
    outfile = tmp_path / f"{uuid.uuid4()}.opl"

    assert 1 == runner('-I', 98, '-o', outfile)
    assert 'Cannot download state information for ID 98.' in caplog.text


def test_change_date_too_old_for_replication_source(runner, tmp_path, replication_server, caplog):
    outfile = tmp_path / f"{uuid.uuid4()}.opl"

    assert 1 == runner('-D', '2015-12-24T08:08:08Z', '-o', outfile)
    assert 'No starting point found' in caplog.text



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.

from npyosmium.osm._osm import ALL as ALL
from npyosmium.osm._osm import AREA as AREA
from npyosmium.osm._osm import CHANGESET as CHANGESET
from npyosmium.osm._osm import NODE as NODE
from npyosmium.osm._osm import NOTHING as NOTHING
from npyosmium.osm._osm import OBJECT as OBJECT
from npyosmium.osm._osm import RELATION as RELATION
from npyosmium.osm._osm import WAY as WAY
from npyosmium.osm._osm import Box as Box
from npyosmium.osm._osm import Location as Location
from npyosmium.osm._osm import osm_entity_bits as osm_entity_bits
from npyosmium.osm.mutable import (
                              create_mutable_node,
                              create_mutable_relation,
                              create_mutable_way,
)
from npyosmium.osm.types import Area as Area
from npyosmium.osm.types import Changeset as Changeset
from npyosmium.osm.types import InnerRing as InnerRing
from npyosmium.osm.types import Node as Node
from npyosmium.osm.types import NodeRef as NodeRef
from npyosmium.osm.types import NodeRefList as NodeRefList
from npyosmium.osm.types import OSMObject as OSMObject
from npyosmium.osm.types import OuterRing as OuterRing
from npyosmium.osm.types import Relation as Relation
from npyosmium.osm.types import RelationMember as RelationMember
from npyosmium.osm.types import RelationMemberList as RelationMemberList
from npyosmium.osm.types import Tag as Tag
from npyosmium.osm.types import TagList as TagList
from npyosmium.osm.types import Way as Way
from npyosmium.osm.types import WayNodeList as WayNodeList

setattr(Location, '__repr__',
        lambda loc: f'npyosmium.osm.Location(x={loc.x!r}, y={loc.y!r})'
                    if loc.valid() else 'npyosmium.osm.Location()')
setattr(Location, '__str__',
        lambda loc: f'{loc.lon_without_check():.7f}/{loc.lat_without_check():.7f}'
                    if loc.valid() else 'invalid')

setattr(Box, '__repr__',
        lambda b: f"npyosmium.osm.Box(bottom_left={b.bottom_left!r}, top_right={b.top_right!r})")
setattr(Box, '__str__', lambda b: f'({b.bottom_left!s} {b.top_right!s})')



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.


def test_member_list_length(simple_handler):
    data = """\
           r2 Mn3@
           r4
           r45 Mw1@fo,r45@4,r45@5
           """

    rels = {}

    def cb(rel):
        rels[rel.id] = len(rel.members)

    simple_handler(data, relation=cb)

    assert rels == {2: 1, 4: 0, 45: 3}


def test_list_members(simple_handler):
    members = []

    def cb(rel):
        members.extend((m.type, m.ref, m.role) for m in rel.members)

    simple_handler("r34 Mn23@,n12@foo,w5@.,r34359737784@(ü)", relation=cb)

    assert members == [('n', 23, ''), ('n', 12, 'foo'), ('w', 5, '.'),
                       ('r', 34359737784, '(ü)')]



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import logging
import time
import uuid
from textwrap import dedent

import pytest
import requests
from helpers import CountingHandler, IDCollector, mkdate
from werkzeug.wrappers import Response

import npyosmium.replication
import npyosmium.replication.server as rserv

pytestmark = [pytest.mark.thread_unsafe, pytest.mark.iterations(1)]


@pytest.mark.parametrize("inp,outp", [
        (None,      'https://text.org/state.txt'),
        (1,         'https://text.org/000/000/001.state.txt'),
        (999,       'https://text.org/000/000/999.state.txt'),
        (1000,      'https://text.org/000/001/000.state.txt'),
        (573923,    'https://text.org/000/573/923.state.txt'),
        (3290012,   'https://text.org/003/290/012.state.txt'),
    ])
def test_get_state_url(inp, outp):
    svr = rserv.ReplicationServer("https://text.org")

    assert outp == svr.get_state_url(inp)


@pytest.mark.parametrize("inp,outp", [
        (1,         'https://who.is/me//000/000/001.osc.gz'),
        (500,       'https://who.is/me//000/000/500.osc.gz'),
        (83750,     'https://who.is/me//000/083/750.osc.gz'),
        (999999999, 'https://who.is/me//999/999/999.osc.gz'),
    ])
def test_get_diff_url(inp, outp):
    svr = rserv.ReplicationServer("https://who.is/me/")

    assert outp, svr.get_diff_url(inp)


def test_get_newest_change_from_file(tmp_path):
    fn = tmp_path / f"{uuid.uuid4()}.opl"
    fn.write_text('n6365 v1 c63965061 t2018-10-29T03:56:07Z i8369524 ux x1 y7')

    val = npyosmium.replication.newest_change_from_file(str(fn))
    assert val == mkdate(2018, 10, 29, 3, 56, 7)


def test_get_state_valid(httpserver):
    httpserver.expect_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z
        txnReadyList=
        txnMax=1219304113
        txnActiveList=1219303583,1219304054,1219304104""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is not None
    assert res.timestamp == mkdate(2017, 8, 26, 11, 4, 2)
    assert res.sequence == 2594669


def test_get_state_sequence_cut(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=259""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is not None
    assert res.timestamp == mkdate(2017, 8, 26, 11, 4, 2)
    assert res.sequence == 2594669


def test_get_state_date_cut(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-2""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is not None
    assert res.timestamp == mkdate(2017, 8, 26, 11, 4, 2)
    assert res.sequence == 2594669


def test_get_state_timestamp_cut(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is not None
    assert res.timestamp == mkdate(2017, 8, 26, 11, 4, 2)
    assert res.sequence == 2594669


def test_get_state_permanent_error(httpserver, caplog):
    httpserver.expect_request('/state.txt').respond_with_data('stuff', status=404)

    with caplog.at_level(logging.DEBUG):
        res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is None
    assert "Loading state info failed" in caplog.text


def test_get_state_transient_error(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data('stuff', status=500)
    httpserver.expect_ordered_request('/state.txt').respond_with_data('stuff', status=500)
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z
        txnReadyList=
        txnMax=1219304113
        txnActiveList=1219303583,1219304054,1219304104""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is not None
    assert res.timestamp == mkdate(2017, 8, 26, 11, 4, 2)
    assert res.sequence == 2594669


def test_get_state_too_many_retries(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=""")
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        #Sat Aug 26 11:04:04 UTC 2017
        txnMaxQueried=1219304113
        sequenceNumber=2594669
        timestamp=2017-08-26T11\\:04\\:02Z""")

    res = rserv.ReplicationServer(httpserver.url_for('')).get_state_info()

    assert res is None


def test_get_state_server_timeout(httpserver):
    def sleeping(request):
        time.sleep(2)
        return Response("""\
    #Sat Aug 26 11:04:04 UTC 2017
    txnMaxQueried=1219304113
    sequenceNumber=2594669
    timestamp=2017-08-26T11\\:04\\:02Z
    txnReadyList=
    txnMax=1219304113
    txnActiveList=1219303583,1219304054,1219304104""")

    httpserver.expect_request("/state.txt").respond_with_handler(sleeping)

    with rserv.ReplicationServer(httpserver.url_for('')) as svr:
        svr.set_request_parameter('timeout', 1)

        res = svr.get_state_info()

    assert res is None


def test_apply_diffs_count(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1
        w1
        r1
    """))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = CountingHandler()
        assert 100 == svr.apply_diffs(h, 100, 10000)
        assert h.counts == [1, 1, 1, 0]


@pytest.mark.parametrize('end_id,max_size, actual_end', [(107, None, 107),
                                                         (None, 512, 108),
                                                         (105, 512, 105),
                                                         (110, 512, 108),
                                                         (None, None, 115)])
def test_apply_diffs_endid(httpserver, end_id, max_size, actual_end):
    httpserver.expect_request('/state.txt').respond_with_data("""\
        sequenceNumber=140
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    for i in range(100, 141):
        httpserver.expect_request(f'/000/000/{i}.opl')\
                  .respond_with_data(f"r{i} M" + ",".join(f"n{i}@" for i in range(1, 3000)))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        res = svr.collect_diffs(101, end_id=end_id, max_size=max_size)

        assert res is not None
        assert res.id == actual_end
        assert res.newest == 140

        ids = IDCollector()
        res.reader.apply(ids)

        assert ids.relations == list(range(101, actual_end + 1))


def test_apply_diffs_without_simplify(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 v23
        n1 v24
        w1
        r1
    """))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = CountingHandler()
        assert 100 == svr.apply_diffs(h, 100, 10000, simplify=False)
        assert [2, 1, 1, 0] == h.counts


def test_apply_diffs_with_simplify(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 v23
        n1 v24
        w1
        r1
    """))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = CountingHandler()
        assert 100 == svr.apply_diffs(h, 100, 10000, simplify=True)
        assert [1, 1, 1, 0] == h.counts


def test_apply_with_location(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 x10.0 y23.0
        w1 Nn1,n2
    """))

    class Handler(CountingHandler):
        def way(self, w):
            self.counts[1] += 1
            assert 2 == len(w.nodes)
            assert 1 == w.nodes[0].ref
            assert 10 == w.nodes[0].location.lon
            assert 23 == w.nodes[0].location.lat
            assert 2 == w.nodes[1].ref
            assert not w.nodes[1].location.valid()

    h = Handler()
    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        assert 100 == svr.apply_diffs(h, 100, 10000, idx="flex_mem")
        assert h.counts == [1, 1, 0, 0]


def test_apply_reader_without_simplify(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 v23
        n1 v24
        w1
        r1
    """))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = CountingHandler()
        diffs = svr.collect_diffs(100, 100000)

    assert diffs is not None
    diffs.reader.apply(h, simplify=False)
    assert [2, 1, 1, 0] == h.counts


def test_apply_reader_with_simplify(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 v23
        n1 v24
        w1
        r1
    """))

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = CountingHandler()
        diffs = svr.collect_diffs(100, 100000)

    assert diffs is not None
    diffs.reader.apply(h, simplify=True)
    assert [1, 1, 1, 0] == h.counts


def test_apply_reader_with_location(httpserver):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 x10.0 y23.0
        w1 Nn1,n2
    """))

    class Handler(CountingHandler):
        def way(self, w):
            self.counts[1] += 1
            assert 2 == len(w.nodes)
            assert 1 == w.nodes[0].ref
            assert 10 == w.nodes[0].location.lon
            assert 23 == w.nodes[0].location.lat
            assert 2 == w.nodes[1].ref
            assert not w.nodes[1].location.valid()

    with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
        h = Handler()
        diffs = svr.collect_diffs(100, 100000)

    assert diffs is not None
    diffs.reader.apply(h, idx="flex_mem")
    assert h.counts == [1, 1, 0, 0]


def test_apply_diffs_permanent_error(httpserver, caplog):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl')\
              .respond_with_data('not a file', status=404)

    with caplog.at_level(logging.ERROR):
        with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
            h = CountingHandler()
            with pytest.raises(requests.HTTPError, match='404'):
                svr.apply_diffs(h, 100, 10000)

    assert 'Permanent server error' in caplog.text


def test_apply_diffs_transient_error_first_diff(httpserver, caplog):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=100
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_request('/000/000/100.opl')\
              .respond_with_data('not a file', status=503)

    with caplog.at_level(logging.ERROR):
        with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
            h = CountingHandler()
            assert svr.apply_diffs(h, 100, 10000) is None
            assert h.counts == [0, 0, 0, 0]

    assert 'Error during diff download' in caplog.text


def test_apply_diffs_permanent_error_later_diff(httpserver, caplog):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=101
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 x10.0 y23.0
        w1 Nn1,n2
    """))
    httpserver.expect_ordered_request('/000/000/101.opl')\
              .respond_with_data('not a file', status=404)

    with caplog.at_level(logging.ERROR):
        with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
            h = CountingHandler()
            assert 100 == svr.apply_diffs(h, 100, 10000)
            assert h.counts == [1, 1, 0, 0]

    assert 'Error during diff download' in caplog.text


def test_apply_diffs_transient_error_later_diff(httpserver, caplog):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=101
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 x10.0 y23.0
        w1 Nn1,n2
    """))
    httpserver.expect_ordered_request('/000/000/101.opl')\
              .respond_with_data('not a file', status=503)
    httpserver.expect_ordered_request('/000/000/101.opl').respond_with_data(dedent("""\
        n2 x10.0 y23.0
    """))

    with caplog.at_level(logging.ERROR):
        with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
            h = CountingHandler()
            assert 101 == svr.apply_diffs(h, 100, 10000)
            assert h.counts == [2, 1, 0, 0]

    assert 'Error during diff download' not in caplog.text


def test_apply_diffs_transient_error_permanent(httpserver, caplog):
    httpserver.expect_ordered_request('/state.txt').respond_with_data("""\
        sequenceNumber=101
        timestamp=2017-08-26T11\\:04\\:02Z
    """)
    httpserver.expect_ordered_request('/000/000/100.opl').respond_with_data(dedent("""\
        n1 x10.0 y23.0
        w1 Nn1,n2
    """))
    for _ in range(4):
        httpserver.expect_ordered_request('/000/000/101.opl')\
                  .respond_with_data('not a file', status=503)
    httpserver.expect_ordered_request('/000/000/101.opl').respond_with_data(dedent("""\
        n2 x10.0 y23.0
    """))

    with caplog.at_level(logging.ERROR):
        with rserv.ReplicationServer(httpserver.url_for(''), "opl") as svr:
            h = CountingHandler()
            assert 100 == svr.apply_diffs(h, 100, 10000)
            assert h.counts == [1, 1, 0, 0]

    assert 'Error during diff download' in caplog.text



"""
Iterate over all ways (and ways only) using node cache to obtain geometries
"""
import sys

import npyosmium


class WayHandler:

    def __init__(self, idx):
        self.idx = idx

    def way(self, w):
        locations = []
        for n in w.nodes:
            locations.append(idx.get(n.ref))  # note that cache is used here
        print(w.id, len(w.nodes), locations)


if len(sys.argv) != 3:
    print("Usage: python use_nodecache.py <osm file> <node cache>")
    exit()

reader = npyosmium.io.Reader(sys.argv[1], npyosmium.osm.osm_entity_bits.WAY)

idx = npyosmium.index.create_map("sparse_file_array," + sys.argv[2])
lh = npyosmium.NodeLocationsForWays(idx)
lh.ignore_errors()

npyosmium.apply(reader, lh, WayHandler(idx))

reader.close()



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
"""
Fetch diffs from an OSM planet server.

The starting point of the diff must be given either as a sequence ID or a date
or can be computed from an OSM file. If no output file is given, the program
will just print the initial sequence ID it would use (or save it in a file, if
requested) and exit. This can be used to bootstrap the update process.

The program tries to download until the latest change on the server is found
or the maximum requested diff size is reached. Note that diffs are kept in
memory during download.

On success, the program will print a single number to stdout, the sequence
number where to continue updates in the next run. This output can also be
written to (and later read from) a file.

*Note:* you may pipe the diff also to standard output using '-o -'. Then
the sequence number will not be printed. You must write it to a file in that
case.

Some OSM data sources require a cookie to be sent with the HTTP requests.
npyosmium-get-changes does not fetch the cookie from these services for you.
However, it can read cookies from a Netscape-style cookie jar file, send these
cookies to the server and will save received cookies to the jar file.
"""
import http.cookiejar
import logging
import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from textwrap import dedent as msgfmt
from typing import List

from .. import SimpleWriter
from ..replication import server as rserv
from ..version import npyosmium_release
from .common import ReplicationStart

log = logging.getLogger()


def write_end_sequence(fname: str, seqid: int) -> None:
    """Either writes out the sequence file or prints the sequence id to stdout.
    """
    if fname is None:
        print(seqid)
    else:
        with open(fname, 'w') as fd:
            fd.write(str(seqid))


def get_arg_parser(from_main: bool = False) -> ArgumentParser:
    parser = ArgumentParser(prog='npyosmium-get-changes',
                            description=__doc__,
                            usage=None if from_main else 'npyosmium-get-changes [options]',
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('-v', dest='loglevel', action='count', default=0,
                        help='Increase verbosity (can be used multiple times)')
    parser.add_argument('-o', '--outfile', dest='outfile',
                        help='Name of diff output file. If omitted, only the '
                             'sequence ID will be printed where updates would start.')
    parser.add_argument('--format', dest='outformat', metavar='FORMAT',
                        help="Format the data should be saved in.")
    parser.add_argument('--server', action='store', dest='server_url',
                        help='Base URL of the replication server')
    parser.add_argument('--diff-type', action='store', dest='server_diff_type', default='osc.gz',
                        help='File format used by the replication server (default: osc.gz)')
    parser.add_argument('--cookie', dest='cookie',
                        help='Netscape-style cookie jar file to read cookies from '
                             'and where received cookies will be written to.')
    parser.add_argument('-s', '--size', dest='outsize', type=int,
                        help='Maximum data to load in MB '
                             '(Defaults to 100MB when no end date/ID has been set).')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-I', '--start-id', dest='start',
                       type=ReplicationStart.from_id, metavar='ID',
                       help='Sequence ID to start with')
    group.add_argument('-D', '--start-date', dest='start', metavar='DATE',
                       type=ReplicationStart.from_date,
                       help='Date when to start updates')
    group.add_argument('-O', '--start-osm-data', dest='start_file', metavar='OSMFILE',
                       help='start at the date of the newest OSM object in the file')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--end-id', dest='end',
                       type=ReplicationStart.from_id, metavar='ID',
                       help='Last sequence ID to download.')
    group.add_argument('-E', '--end-date', dest='end', metavar='DATE',
                       type=ReplicationStart.from_date,
                       help='Do not download diffs later than the given date.')
    parser.add_argument('-f', '--sequence-file', dest='seq_file',
                        help='Sequence file. If the file exists, then updates '
                             'will start after the id given in the file. At the '
                             'end of the process, the last sequence ID contained '
                             'in the diff is written.')
    parser.add_argument('--ignore-osmosis-headers', dest='ignore_headers',
                        action='store_true',
                        help='When determining the start from an OSM file, '
                             'ignore potential replication information in the '
                             'header and search for the newest OSM object.')
    parser.add_argument('-d', '--no-deduplicate', action='store_false', dest='simplify',
                        help='Do not deduplicate diffs.')
    parser.add_argument('--socket-timeout', dest='socket_timeout', type=int, default=60,
                        help='Set timeout for file downloads.')
    parser.add_argument('--version', action='version', version='npyosmium ' + npyosmium_release)

    return parser


def pyosmium_get_changes(args: List[str]) -> int:
    logging.basicConfig(stream=sys.stderr,
                        format='%(asctime)s %(levelname)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')

    options = get_arg_parser(from_main=True).parse_args(args)

    log.setLevel(max(3 - options.loglevel, 0) * 10)

    if options.outfile == '-' and options.outformat is None:
        log.error("You must define a format when using stdout. See --format.")
        return 1

    if options.start_file is not None:
        options.start = ReplicationStart.from_osm_file(options.start_file,
                                                       options.ignore_headers)
    if options.start is None:
        if options.seq_file is None:
            log.error(msgfmt("""
              Don't know with which change to start. One of the parameters
                -I / -D / -O / -f
              needs to be given."""))
            return 1

        with open(options.seq_file, 'r') as f:
            seq = f.readline()
            options.start = ReplicationStart.from_id(seq)

    if options.server_url is not None and options.start.source is not None:
        if options.server_url != options.start.source:
            log.error(msgfmt(f"""
                You asked to use server URL:
                  {options.server_url}
                but the referenced OSM file points to replication server:
                  {options.start.source}
                If you really mean to overwrite the URL, use --ignore-osmosis-headers."""))
            return 2
    url = options.server_url \
        or options.start.source \
        or 'https://planet.osm.org/replication/minute/'
    logging.info("Using replication server at %s" % url)

    with rserv.ReplicationServer(url, diff_type=options.server_diff_type) as svr:
        svr.set_request_parameter('timeout', options.socket_timeout or None)

        if options.cookie is not None:
            cookie_jar = http.cookiejar.MozillaCookieJar(options.cookie)
            cookie_jar.load(options.cookie)
            svr.set_request_parameter('cookies', cookie_jar)

        # Sanity check if server URL is correct and server is responding.
        current = svr.get_state_info()
        if current is None:
            log.error("Cannot download state information. Is the replication URL correct?")
            return 3
        log.debug(f"Server is at sequence {current.sequence} ({current.timestamp}).")

        startseq = options.start.get_sequence(svr)
        if startseq is None:
            log.error(f"No starting point found for time {options.start.date} on server {url}")
            return 1

        if options.outfile is None:
            write_end_sequence(options.seq_file, startseq - 1)
            return 0

        log.debug("Starting download at ID %d (max %f MB)"
                  % (startseq, options.outsize or float('inf')))
        if options.outformat is not None:
            outhandler = SimpleWriter(options.outfile, filetype=options.outformat)
        else:
            outhandler = SimpleWriter(options.outfile)

        if options.outsize is not None:
            max_size = options.outsize * 1024
        elif options.end is None:
            max_size = 100 * 1024
        else:
            max_size = None

        if options.end is None:
            end_id = None
        else:
            end_id = options.end.get_end_sequence(svr)
            if end_id is None:
                log.error("Cannot find the end date/ID on the server.")
                return 1

        endseq = svr.apply_diffs(outhandler, startseq, max_size=max_size,
                                 end_id=end_id, simplify=options.simplify)
        outhandler.close()

    # save cookies
    if options.cookie:
        cookie_jar.save(options.cookie)

    if endseq is None:
        log.error("Error while downloading diffs.")
        return 3

    if options.outfile != '-' or options.seq_file is not None:
        write_end_sequence(options.seq_file, endseq)

    return 0


def main() -> int:
    logging.basicConfig(stream=sys.stderr,
                        format='%(asctime)s %(levelname)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')

    return pyosmium_get_changes(sys.argv[1:])



import sys

import npyosmium

if len(sys.argv) != 3:
    print("Usage: python create_nodecache.py <osm file> <node cache>")
    exit(-1)

reader = npyosmium.io.Reader(sys.argv[1], npyosmium.osm.osm_entity_bits.NODE)

idx = npyosmium.index.create_map("sparse_file_array," + sys.argv[2])
lh = npyosmium.NodeLocationsForWays(idx)

npyosmium.apply(reader, lh)

reader.close()



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2023 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
"""
Version information.
"""

# current release (Pip version)
npyosmium_release = '3.7.0'



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import uuid

import pytest
from helpers import CountingHandler

import npyosmium


class NullHandler:

    def node(self, n):
        pass


def _run_file(fn):
    with npyosmium.io.Reader(fn) as rd:
        npyosmium.apply(rd, NullHandler())


@pytest.mark.parametrize('as_string', [True, False])
def test_file_simple(tmp_path, as_string):
    fn = tmp_path / f"{uuid.uuid4()}.opl"
    fn.write_text('n1')

    if as_string:
        fn = str(fn)

    for n in npyosmium.FileProcessor(npyosmium.io.File(fn)):
        assert n.is_node()
        assert n.id == 1


@pytest.mark.parametrize('as_string', [True, False])
def test_file_with_format(tmp_path, as_string):
    fn = tmp_path / f"{uuid.uuid4()}.txt"
    fn.write_text('n1')

    if as_string:
        fn = str(fn)

    for n in npyosmium.FileProcessor(npyosmium.io.File(fn, 'opl')):
        assert n.is_node()
        assert n.id == 1


def test_node_only(test_data):
    _run_file(test_data('n1'))


def test_way_only(test_data):
    _run_file(test_data('w1 Nn1,n2,n3'))


def test_relation_only(test_data):
    _run_file(test_data('r573 Mw1@'))


def test_node_with_tags(test_data):
    _run_file(test_data('n32 Tbar=xx'))


def test_way_with_tags(test_data):
    _run_file(test_data('w5666 Nn1,n2,n3 Tbar=xx'))


def test_relation_with_tags(test_data):
    _run_file(test_data('r573 Mw1@ Tbar=xx'))


def test_broken_timestamp(test_data):
    fn = test_data('n1 tx')

    with npyosmium.io.Reader(fn) as rd:
        with pytest.raises(RuntimeError):
            npyosmium.apply(rd, NullHandler())


@pytest.mark.parametrize('as_string', [True, False])
def test_file_header(tmp_path, as_string):
    fn = tmp_path / f"{uuid.uuid4()}.xml"
    fn.write_text("""<?xml version='1.0' encoding='UTF-8'?>
    <osm version="0.6" generator="test-pyosmium" timestamp="2014-08-26T20:22:02Z">
         <bounds minlat="-90" minlon="-180" maxlat="90" maxlon="180"/>
    </osm>
    """)

    if as_string:
        fn = str(fn)

    with npyosmium.io.Reader(fn) as rd:
        h = rd.header()
        assert not h.has_multiple_object_versions
        assert h.box().valid()
        assert h.box().size() == 64800.0


def test_reader_with_filebuffer():
    rd = npyosmium.io.Reader(npyosmium.io.FileBuffer('n1 x4 y1'.encode('utf-8'), 'opl'))
    handler = CountingHandler()

    npyosmium.apply(rd, handler)

    assert handler.counts == [1, 0, 0, 0]



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Union

from npyosmium import IdTracker
from npyosmium._osmium import SimpleWriter
from npyosmium.file_processor import FileProcessor, zip_processors
from npyosmium.io import File, FileBuffer



class BackReferenceWriter:
    """ Writer that adds referenced objects, so that all written
        objects are reference-complete.

        The collected data is first written into a temporary file and
        the necessary references are tracked internally. When the writer
        is closed, it writes the final file, mixing together the referenced
        objects from the original file and the written data.

        The writer should usually be used as a context manager.
    """

    def __init__(self, outfile: Union[str, 'os.PathLike[str]', File],
                 ref_src: Union[str, 'os.PathLike[str]', File, FileBuffer],
                 overwrite: bool = False, remove_tags: bool = True,
                 relation_depth: int = 0):
        """ Create a new writer.

            `outfile` is the name of the output file to write. The file must
            not yet exist unless `overwrite` is set to True.

            `ref_src` is the OSM input file, where to take the reference objects
            from. This is usually the same file the data to be written is taken
            from.

            The writer will by default remove all tags from referenced objects,
            so that they do not appear as stray objects in the file. Set
            `remove_tags` to False to keep the tags.

            The writer will not complete nested relations by default. If you
            need nested relations, set `relation_depth` to the minimum depth
            to which relations shall be completed.
        """
        self.outfile = outfile
        self.tmpdir = TemporaryDirectory()
        self.writer = SimpleWriter(str(Path(self.tmpdir.name, 'back_writer.osm.pbf')))
        self.overwrite = overwrite
        self.remove_tags = remove_tags
        self.id_tracker = IdTracker()
        self.ref_src = ref_src
        self.relation_depth = relation_depth

    def __enter__(self) -> 'BackReferenceWriter':
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        if exc_type is None:
            self.close()
        else:
            # Exception occured. Do not write out the final file.
            self.writer.close()

    def add(self, obj: Any) -> None:
        """ Write an arbitrary OSM object. This can be either an
            osmium object or a Python object that has the appropriate
            attributes.
        """
        self.id_tracker.add_references(obj)
        self.writer.add(obj)

    def add_node(self, n: Any) -> None:
        """ Write out an OSM node.
        """
        self.writer.add_node(n)

    def add_way(self, w: Any) -> None:
        """ Write out an OSM way.
        """
        self.id_tracker.add_references(w)
        self.writer.add_way(w)

    def add_relation(self, r: Any) -> None:
        """ Write out an OSM relation.
        """
        self.id_tracker.add_references(r)
        self.writer.add_relation(r)

    def close(self) -> None:
        """ Close the writer and write out the final file.

            The function will be automatically called when the writer
            is used as a context manager.
        """
        self.writer.close()
        self.id_tracker.complete_backward_references(self.ref_src,
                                                     relation_depth=self.relation_depth)

        fp1 = FileProcessor(str(Path(self.tmpdir.name, 'back_writer.osm.pbf')))
        fp2 = FileProcessor(self.ref_src).with_filter(self.id_tracker.id_filter())

        with SimpleWriter(self.outfile, overwrite=self.overwrite) as writer:
            for o1, o2 in zip_processors(fp1, fp2):
                if o1:
                    writer.add(o1)
                elif o2:
                    if self.remove_tags and hasattr(o2, 'replace'):
                        writer.add(o2.replace(tags={}))
                    else:
                        writer.add(o2)

        self.tmpdir.cleanup()



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import Optional
import logging
from dataclasses import dataclass
import datetime as dt
from argparse import ArgumentTypeError

from ..replication import newest_change_from_file
from ..replication.server import ReplicationServer
from ..replication.utils import get_replication_header


log = logging.getLogger()


@dataclass
class ReplicationStart:
    """ Represents the point where changeset download should begin.
    """
    date: Optional[dt.datetime] = None
    seq_id: Optional[int] = None
    source: Optional[str] = None

    def get_sequence(self, svr: ReplicationServer) -> Optional[int]:
        if self.seq_id is not None:
            log.debug("Using given sequence ID %d" % self.seq_id)
            if self.seq_id > 0:
                start_state = svr.get_state_info(seq=self.seq_id)
                if start_state is None:
                    log.error(
                        f"Cannot download state information for ID {self.seq_id}."
                        " Server may not have this diff anymore.")
                    return None
                self.date = start_state.timestamp
            return self.seq_id + 1

        assert self.date is not None
        log.debug("Looking up sequence ID for timestamp %s" % self.date)
        return svr.timestamp_to_sequence(self.date, limit_by_oldest_available=True)

    def get_end_sequence(self, svr: ReplicationServer) -> Optional[int]:
        if self.seq_id is not None:
            log.debug("Using end sequence ID %d" % self.seq_id)
            return self.seq_id

        assert self.date is not None
        log.debug("Looking up end sequence ID for timestamp %s" % self.date)
        return svr.timestamp_to_sequence(self.date)

    @staticmethod
    def from_id(idstr: str) -> 'ReplicationStart':
        try:
            seq_id = int(idstr)
        except ValueError:
            raise ArgumentTypeError("Sequence id '%s' is not a number" % idstr)

        if seq_id < -1:
            raise ArgumentTypeError("Sequence id '%s' is negative" % idstr)

        return ReplicationStart(seq_id=seq_id)

    @staticmethod
    def from_date(datestr: str) -> 'ReplicationStart':
        try:
            date = dt.datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%SZ")
            date = date.replace(tzinfo=dt.timezone.utc)
        except ValueError:
            raise ArgumentTypeError(
                "Date needs to be in ISO8601 format (e.g. 2015-12-24T08:08:08Z).")

        return ReplicationStart(date=date)

    @staticmethod
    def from_osm_file(fname: str, ignore_headers: bool) -> 'ReplicationStart':
        if ignore_headers:
            ts = None
            seq = None
            url = None
        else:
            try:
                (url, seq, ts) = get_replication_header(fname)
            except RuntimeError as e:
                raise ArgumentTypeError(e)

        if ts is None and seq is None:
            log.debug("OSM file has no replication headers. Looking for newest OSM object.")
            try:
                ts = newest_change_from_file(fname)
            except RuntimeError as e:
                raise ArgumentTypeError(e)

            if ts is None:
                raise ArgumentTypeError("OSM file does not seem to contain valid data.")

        return ReplicationStart(seq_id=seq, date=ts, source=url)



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import npyosmium


def test_set_unset_empty():
    ids = npyosmium.index.IdSet()
    assert ids.empty()
    assert not ids

    ids.set(34)
    assert not ids.empty()
    assert ids

    ids.unset(34)
    assert ids.empty()
    assert not ids


def test_set_get():
    ids = npyosmium.index.IdSet()

    for i in (1, 100, 2):
        ids.set(i)

    assert ids.get(100)
    assert not ids.get(3)
    assert 1 in ids
    assert 45 not in ids


def test_clear_and_size():
    ids = npyosmium.index.IdSet()

    assert len(ids) == 0

    for i in (1, 100, 2):
        ids.set(i)

    assert len(ids) == 3

    ids.clear()

    assert len(ids) == 0



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import ClassVar, Tuple, Optional

from typing import overload
import datetime

import npyosmium.osm.mutable
import npyosmium.osm.types

ALL: osm_entity_bits
AREA: osm_entity_bits
CHANGESET: osm_entity_bits
NODE: osm_entity_bits
NOTHING: osm_entity_bits
OBJECT: osm_entity_bits
RELATION: osm_entity_bits
WAY: osm_entity_bits


class Location:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, lon: float, lat: float) -> None: ...
    def lat_without_check(self) -> float: ...
    def lon_without_check(self) -> float: ...
    def valid(self) -> bool: ...
    @property
    def lat(self) -> float: ...
    @property
    def lon(self) -> float: ...
    @property
    def x(self) -> int: ...
    @property
    def y(self) -> int: ...

class Box:
    @overload
    def __init__(self, minx: float, miny: float, maxx: float, maxy: float) -> None: ...
    @overload
    def __init__(self, bottom_left: Location, top_right: Location) -> None: ...
    def contains(self, location: Location) -> bool: ...
    @overload
    def extend(self, location: Location) -> Box: ...
    @overload
    def extend(self, box: Box) -> Box: ...
    def size(self) -> float: ...
    def valid(self) -> bool: ...
    @property
    def bottom_left(self) -> Location: ...
    @property
    def top_right(self) -> Location: ...

class CTagListIterator:
    pass

class CRelationMemberListIterator:
    pass

class BufferProxyProtocol:
    def is_valid(self) -> bool: ...

class NodeRefList:
    def ends_have_same_location(self, parent: BufferProxyProtocol) -> bool: ...
    def is_closed(self, parent: BufferProxyProtocol) -> bool: ...
    def get(self, parent: BufferProxyProtocol, idx: int) -> npyosmium.osm.types.NodeRef: ...
    def size(self, parent: BufferProxyProtocol) -> int: ...

class CWayNodeList(NodeRefList):
    pass

class COuterRing(NodeRefList):
    pass

class CInnerRing(NodeRefList):
    pass

class COuterRingIterator:
    pass

class CInnerRingIterator:
    pass

class TagContainerProtocol(BufferProxyProtocol):
    def tags_size(self) -> int: ...
    def tags_get_value_by_key(self, key: str, default: Optional[str]) -> Optional[str]: ...
    def tags_has_key(self, key: str) -> bool: ...
    def tags_begin(self) -> CTagListIterator: ...
    def tags_next(self, it: CTagListIterator) -> npyosmium.osm.types.Tag: ...

class COSMObject(TagContainerProtocol):
    def positive_id(self) -> int: ...
    def user_is_anonymous(self) -> bool: ...
    def changeset(self) -> int: ...
    def deleted(self) -> bool: ...
    def id(self) -> int: ...
    def timestamp(self) -> datetime.datetime: ...
    def uid(self) -> int: ...
    def user(self) -> str: ...
    def version(self) -> int: ...
    def visible(self) -> bool: ...

class COSMNode(COSMObject):
    def location(self) -> Location: ...

class COSMWay(COSMObject):
    def ends_have_same_id(self) -> bool: ...
    def ends_have_same_location(self) -> bool: ...
    def is_closed(self) -> bool: ...
    def nodes(self) -> CWayNodeList: ...

class COSMRelation(COSMObject):
    def members_size(self) -> int: ...
    def members_begin(self) -> CRelationMemberListIterator: ...
    def members_next(self, it: CRelationMemberListIterator) -> npyosmium.osm.types.RelationMember: ...

class COSMArea(COSMObject):
    def from_way(self) -> bool: ...
    def inner_rings(self, outer_ring: COuterRing) -> CInnerRingIterator: ...
    def is_multipolygon(self) -> bool: ...
    def num_rings(self) -> Tuple[int,int]: ...
    def orig_id(self) -> int: ...
    def outer_begin(self) -> COuterRingIterator: ...
    def outer_next(self, it: COuterRingIterator) -> COuterRing: ...
    def inner_begin(self, oring: COuterRing) -> CInnerRingIterator: ...
    def inner_next(self, it: CInnerRingIterator) -> CInnerRing: ...

class COSMChangeset(TagContainerProtocol):
    def user_is_anonymous(self) -> bool: ...
    def bounds(self) -> Box: ...
    def closed_at(self) -> datetime.datetime: ...
    def created_at(self) -> datetime.datetime: ...
    def id(self) -> int: ...
    def num_changes(self) -> int: ...
    def open(self) -> bool: ...
    def uid(self) -> int: ...
    def user(self) -> str: ...

class osm_entity_bits:
    ALL: ClassVar[osm_entity_bits] = ...
    AREA: ClassVar[osm_entity_bits] = ...
    CHANGESET: ClassVar[osm_entity_bits] = ...
    NODE: ClassVar[osm_entity_bits] = ...
    NOTHING: ClassVar[osm_entity_bits] = ...
    OBJECT: ClassVar[osm_entity_bits] = ...
    RELATION: ClassVar[osm_entity_bits] = ...
    WAY: ClassVar[osm_entity_bits] = ...
    def __init__(self, value: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...
    def __or__(self, other: osm_entity_bits) -> osm_entity_bits: ...
    def __and__(self, other: osm_entity_bits) -> osm_entity_bits: ...
    def __invert__(self) -> osm_entity_bits: ...
    def __bool__(self) -> bool: ...



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import Union, List, TYPE_CHECKING

if TYPE_CHECKING:
    import os
    from typing_extensions import Buffer
    from ._osmium import HandlerLike

from ._osmium import apply, NodeLocationsForWays
from .io import Reader, File, FileBuffer
from .osm import osm_entity_bits
from .area import AreaManager
from .index import create_map


class SimpleHandler:
    """ The most generic of OSM data handlers. Derive your data processor
        from this class and implement callbacks for each object type you are
        interested in. The following data types are recognised:

          `node`, `way`, `relation`, `area` and `changeset`

        A callback takes exactly one parameter which is the object. Note that
        all objects that are handed into the handler are only readable and are
        only valid until the end of the callback is reached. Any data that
        should be retained must be copied into other data structures.
    """
    def enabled_for(self) -> osm_entity_bits:
        """ Return the list of OSM object types this handler will handle.
        """
        entities = osm_entity_bits.NOTHING
        if hasattr(self, 'node'):
            entities |= osm_entity_bits.NODE
        if hasattr(self, 'way'):
            entities |= osm_entity_bits.WAY
        if hasattr(self, 'relation'):
            entities |= osm_entity_bits.RELATION
        if hasattr(self, 'area'):
            entities |= osm_entity_bits.AREA
        if hasattr(self, 'changeset'):
            entities |= osm_entity_bits.CHANGESET

        return entities

    def apply_file(self, filename: Union[str, 'os.PathLike[str]', File],
                   locations: bool = False, idx: str = 'flex_mem',
                   filters: List['HandlerLike'] = []) -> None:
        """ Apply the handler to the given file. If locations is true, then
            a location handler will be applied before, which saves the node
            positions. In that case, the type of this position index can be
            further selected in idx. If an area callback is implemented, then
            the file will be scanned twice and a location handler and a
            handler for assembling multipolygons and areas from ways will
            be executed.
        """
        self._apply_object(filename, locations, idx, filters)

    def apply_buffer(self, buffer: 'Buffer', format: str,
                     locations: bool = False, idx: str = 'flex_mem',
                     filters: List['HandlerLike'] = []) -> None:
        """Apply the handler to a string buffer. The buffer must be a
           byte string.
        """
        self._apply_object(FileBuffer(buffer, format), locations, idx, filters)

    def _apply_object(self, obj: Union[str, 'os.PathLike[str]', File, FileBuffer],
                      locations: bool, idx: str,
                      filters: List['HandlerLike']) -> None:
        entities = self.enabled_for()
        if entities & osm_entity_bits.AREA:
            area = AreaManager()
            with Reader(obj, osm_entity_bits.RELATION) as rd:
                apply(rd, *filters, area.first_pass_handler())

            entities |= osm_entity_bits.OBJECT
            lh = NodeLocationsForWays(create_map(idx))
            lh.ignore_errors()
            handlers = [lh, area.second_pass_handler(*filters, self), *filters, self]
        elif locations:
            entities |= osm_entity_bits.NODE
            lh = NodeLocationsForWays(create_map(idx))
            lh.ignore_errors()
            handlers = [lh, *filters, self]
        else:
            handlers = [*filters, self]

        with Reader(obj, entities) as rd:
            apply(rd, *handlers)



"""
Filter all objects with a coastline tag.

This example shows how to write objects to a file.

We need to go twice over the file. First read the ways, filter the ones
we are interested in and remember the nodes required. Then, in a second
run all the relevant nodes and ways are written out.
"""

import sys

import npyosmium

if __name__ == '__main__':
    if len(sys.argv) != 3:
        print("Usage: python filter_coastlines.py <infile> <outfile>")
        sys.exit(-1)

    # go through the ways to find all relevant nodes
    nodes = set()
    # Pre-filter the ways by tags. The less object we need to look at, the better.
    way_filter = npyosmium.filter.KeyFilter('natural')
    # only scan the ways of the file
    for obj in npyosmium.FileProcessor(sys.argv[1], npyosmium.osm.WAY).with_filter(way_filter):
        if obj.tags['natural'] == 'coastline':
            nodes.update(n.ref for n in obj.nodes)

    # go through the file again and write out the data
    writer = npyosmium.SimpleWriter(sys.argv[2])

    # This time the pre-filtering should only apply to ways.
    way_filter = npyosmium.filter.KeyFilter('natural').enable_for(npyosmium.osm.WAY)

    # We need nodes and ways in the second pass.
    for obj in npyosmium.FileProcessor(sys.argv[1], npyosmium.osm.WAY | npyosmium.osm.NODE) \
                     .with_filter(way_filter):
        if obj.is_node() and obj.id in nodes:
            # Strip the object of tags along the way
            writer.add_node(obj.replace(tags={}))
        elif obj.is_way() and obj.tags['natural'] == 'coastline':
            writer.add_way(obj)

    writer.close()



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
from typing import List

import npyosmium.osm

class LocationTable:
    """ A map from a node ID to a location object. Location can be set
        and queried using the standard [] notation for dicts.
        This implementation works only with positive node IDs.
    """
    def clear(self) -> None:
        """ Remove all entries from the location table..
        """
    def get(self, id: int) -> npyosmium.osm.Location:
        """ Get the location for the given node ID. Raises a `KeyError`
            when there is no location for the given id.
        """
    def set(self, id: int, loc: npyosmium.osm.Location) -> None:
        """ Set the location for the given node ID.
        """
    def used_memory(self) -> int:
        """ Return the size (in bytes) currently allocated by this location table.
        """
    def __getitem__(self, id: int) -> npyosmium.osm.Location: ...
    def __setitem__(self, id: int, loc: npyosmium.osm.Location) -> None: ...



class IdSet:
    """ Compact storage for a set of IDs.
    """
    def __init__(self) -> None:
        """ Initialise an empty set.
        """
    def set(self, id: int) -> None:
        """ Add an ID to the storage.
            Does nothing if the ID is already contained.
        """
    def unset(self, id: int) -> None:
        """ Remove an ID from the storage.
            Does nothing if the ID is not in the storage.
        """
    def get(self, id: int) -> bool:
        """ Check if the given ID is in the storage.
        """
    def empty(self) -> bool:
        """ Check if no IDs are stored yet.
        """
    def clear(self) -> None:
        """ Remove all IDs from the set.
        """
    def __len__(self) -> int: ...
    def __contains__(self, id: int) -> bool: ...


def create_map(map_type: str) -> LocationTable:
    """ Create a new location store. Use the _map_type_ parameter
        to choose a concrete implementation. Some implementations
        take additiona configuration parameters, which can also be
        set through the _map_type_ argument. For example,
        to create an array cache backed by a file 'foo.store',
        the map_type needs to be set to `dense_file_array,foo.store`.
        Read the section on [location storage][location-storage] in the
        user manual for more information about the different implementations.
    """
def map_types() -> List[str]:
    """ Return a list of strings with valid types for the location table.
    """



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2024 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import datetime

def newest_change_from_file(filename: str) -> datetime.datetime:
    """ Find the data of the most recent change in a file.
    """



# SPDX-License-Identifier: BSD-2-Clause
#
# This file is part of pyosmium. (https://osmcode.org/pyosmium/)
#
# Copyright (C) 2025 Sarah Hoffmann <lonvia@denofr.de> and others.
# For a full list of authors see the git log.
import uuid
from pathlib import Path
from textwrap import dedent

import pytest

import npyosmium


@pytest.fixture
def test_data_dir():
    return (Path(__file__) / '..').resolve()


@pytest.fixture
def to_opl():
    def _make(data):
        if isinstance(data, (list, tuple)):
            return '\n'.join(data)

        return dedent(data)

    return _make


@pytest.fixture
def test_data(tmp_path, to_opl):

    def _mkfile(data):
        filename = tmp_path / (str(uuid.uuid4()) + '.opl')
        filename.write_text(to_opl(data))
        return str(filename)

    return _mkfile


@pytest.fixture
def opl_buffer(to_opl):

    def _mkbuffer(data):
        return npyosmium.io.FileBuffer(to_opl(data).encode('utf-8'), 'opl')

    return _mkbuffer


@pytest.fixture
def opl_reader(opl_buffer):

    def _mkbuffer(data):
        return npyosmium.io.Reader(opl_buffer(data))

    return _mkbuffer


@pytest.fixture
def simple_handler(to_opl):

    def _run(data, node=None, way=None, relation=None, area=None, locations=False):
        handler = npyosmium.make_simple_handler(node=node, way=way, relation=relation, area=area)
        handler.apply_buffer(to_opl(data).encode('utf-8'), 'opl', locations=locations)

    return _run
